@@ -105,6 +105,16 @@ class MappingException extends \Doctrine\ORM\ORMException
         return new self("The association mapping '$fieldName' misses the 'sourceEntity' attribute.");
     }
 
+    /**
+     * @param string $fieldName
+     *
+     * @return MappingException
+     */
+    public static function missingEmbeddedClass($fieldName)
+    {
+        return new self("The embed mapping '$fieldName' misses the 'class' attribute.");
+    }
+
     /**
      * @param string $entityName
      * @param string $fileName
@@ -655,7 +665,7 @@ class MappingException extends \Doctrine\ORM\ORMException
      */
     public static function noInheritanceOnMappedSuperClass($className)
     {
-        return new self("Its not supported to define inheritance information on a mapped superclass '" . $className . "'.");
+        return new self("It is not supported to define inheritance information on a mapped superclass '" . $className . "'.");
     }
 
     /**
@@ -707,6 +717,18 @@ class MappingException extends \Doctrine\ORM\ORMException
         return new self(sprintf('Entity Listener "%s" declared on "%s" has no method "%s".', $listenerName, $className, $methodName));
     }
 
+    /**
+     * @param string $listenerName
+     * @param string $methodName
+     * @param string $className
+     *
+     * @return \Doctrine\ORM\Mapping\MappingException
+     */
+    public static function duplicateEntityListener($listenerName, $methodName, $className)
+    {
+        return new self(sprintf('Entity Listener "%s#%s()" in "%s" was already declared, but it must be declared only once.', $listenerName, $methodName, $className));
+    }
+
     /**
      * @param string $className
      * @param string $annotation
@@ -769,4 +791,22 @@ class MappingException extends \Doctrine\ORM\ORMException
             sprintf('Missing "sequenceName" attribute for sequence id generator definition on class "%s".', $className)
         );
     }
+
+    /**
+     * @param string $className
+     * @param string $propertyName
+     *
+     * @return MappingException
+     */
+    public static function infiniteEmbeddableNesting($className, $propertyName)
+    {
+        return new self(
+            sprintf(
+                'Infinite nesting detected for embedded property %s::%s. ' .
+                'You cannot embed an embeddable from the same type inside an embeddable.',
+                $className,
+                $propertyName
+            )
+        );
+    }
 }
