@@ -20,15 +20,14 @@
 namespace Doctrine\ORM\Tools;
 
 use Doctrine\ORM\ORMException;
-use Doctrine\DBAL\Types\Type;
 use Doctrine\DBAL\Schema\Comparator;
+use Doctrine\DBAL\Schema\Index;
 use Doctrine\DBAL\Schema\Schema;
 use Doctrine\DBAL\Schema\Table;
 use Doctrine\DBAL\Schema\Visitor\DropSchemaSqlCollector;
 use Doctrine\DBAL\Schema\Visitor\RemoveNamespacedAssets;
 use Doctrine\ORM\EntityManagerInterface;
 use Doctrine\ORM\Mapping\ClassMetadata;
-use Doctrine\ORM\Internal\CommitOrderCalculator;
 use Doctrine\ORM\Tools\Event\GenerateSchemaTableEventArgs;
 use Doctrine\ORM\Tools\Event\GenerateSchemaEventArgs;
 
@@ -126,6 +125,7 @@ class SchemaTool
         return (
             isset($processedClasses[$class->name]) ||
             $class->isMappedSuperclass ||
+            $class->isEmbeddedClass ||
             ($class->isInheritanceTypeSingleTable() && $class->name != $class->rootEntityName)
         );
     }
@@ -265,13 +265,26 @@ class SchemaTool
 
             if (isset($class->table['indexes'])) {
                 foreach ($class->table['indexes'] as $indexName => $indexData) {
-                    $table->addIndex($indexData['columns'], is_numeric($indexName) ? null : $indexName);
+                    if( ! isset($indexData['flags'])) {
+                        $indexData['flags'] = array();
+                    }
+
+                    $table->addIndex($indexData['columns'], is_numeric($indexName) ? null : $indexName, (array)$indexData['flags'], isset($indexData['options']) ? $indexData['options'] : array());
                 }
             }
 
             if (isset($class->table['uniqueConstraints'])) {
                 foreach ($class->table['uniqueConstraints'] as $indexName => $indexData) {
-                    $table->addUniqueIndex($indexData['columns'], is_numeric($indexName) ? null : $indexName);
+                    $uniqIndex = new Index($indexName, $indexData['columns'], true, false, [], isset($indexData['options']) ? $indexData['options'] : []);
+
+                    foreach ($table->getIndexes() as $tableIndexName => $tableIndex) {
+                        if ($tableIndex->isFullfilledBy($uniqIndex)) {
+                            $table->dropIndex($tableIndexName);
+                            break;
+                        }
+                    }
+
+                    $table->addUniqueIndex($indexData['columns'], is_numeric($indexName) ? null : $indexName, isset($indexData['options']) ? $indexData['options'] : array());
                 }
             }
 
@@ -332,7 +345,7 @@ class SchemaTool
         $discrColumn = $class->discriminatorColumn;
 
         if ( ! isset($discrColumn['type']) ||
-            (strtolower($discrColumn['type']) == 'string' && $discrColumn['length'] === null)
+            (strtolower($discrColumn['type']) == 'string' && ! isset($discrColumn['length']))
         ) {
             $discrColumn['type'] = 'string';
             $discrColumn['length'] = 255;
@@ -484,7 +497,7 @@ class SchemaTool
             $foreignClass = $this->em->getClassMetadata($mapping['targetEntity']);
 
             if ($mapping['type'] & ClassMetadata::TO_ONE && $mapping['isOwningSide']) {
-                $primaryKeyColumns = $uniqueConstraints = array(); // PK is unnecessary for this relation-type
+                $primaryKeyColumns = array(); // PK is unnecessary for this relation-type
 
                 $this->gatherRelationJoinColumns(
                     $mapping['joinColumns'],
@@ -492,14 +505,9 @@ class SchemaTool
                     $foreignClass,
                     $mapping,
                     $primaryKeyColumns,
-                    $uniqueConstraints,
                     $addedFks,
                     $blacklistedFks
                 );
-
-                foreach ($uniqueConstraints as $indexName => $unique) {
-                    $table->addUniqueIndex($unique['columns'], is_numeric($indexName) ? null : $indexName);
-                }
             } elseif ($mapping['type'] == ClassMetadata::ONE_TO_MANY && $mapping['isOwningSide']) {
                 //... create join table, one-many through join table supported later
                 throw ORMException::notSupported();
@@ -511,7 +519,7 @@ class SchemaTool
                     $this->quoteStrategy->getJoinTableName($mapping, $foreignClass, $this->platform)
                 );
 
-                $primaryKeyColumns = $uniqueConstraints = array();
+                $primaryKeyColumns = array();
 
                 // Build first FK constraint (relation table => source table)
                 $this->gatherRelationJoinColumns(
@@ -520,7 +528,6 @@ class SchemaTool
                     $class,
                     $mapping,
                     $primaryKeyColumns,
-                    $uniqueConstraints,
                     $addedFks,
                     $blacklistedFks
                 );
@@ -532,16 +539,11 @@ class SchemaTool
                     $foreignClass,
                     $mapping,
                     $primaryKeyColumns,
-                    $uniqueConstraints,
                     $addedFks,
                     $blacklistedFks
                 );
 
                 $theJoinTable->setPrimaryKey($primaryKeyColumns);
-
-                foreach ($uniqueConstraints as $indexName => $unique) {
-                    $theJoinTable->addUniqueIndex($unique['columns'], is_numeric($indexName) ? null : $indexName);
-                }
             }
         }
     }
@@ -592,7 +594,6 @@ class SchemaTool
      * @param ClassMetadata $class
      * @param array         $mapping
      * @param array         $primaryKeyColumns
-     * @param array         $uniqueConstraints
      * @param array         $addedFks
      * @param array         $blacklistedFks
      *
@@ -606,7 +607,6 @@ class SchemaTool
         $class,
         $mapping,
         &$primaryKeyColumns,
-        &$uniqueConstraints,
         &$addedFks,
         &$blacklistedFks
     ) {
@@ -614,6 +614,7 @@ class SchemaTool
         $foreignColumns     = array();
         $fkOptions          = array();
         $foreignTableName   = $this->quoteStrategy->getTableName($class, $this->platform);
+        $uniqueConstraints  = array();
 
         foreach ($joinColumns as $joinColumn) {
 
@@ -683,6 +684,12 @@ class SchemaTool
             }
         }
 
+        // Prefer unique constraints over implicit simple indexes created for foreign keys.
+        // Also avoids index duplication.
+        foreach ($uniqueConstraints as $indexName => $unique) {
+            $theJoinTable->addUniqueIndex($unique['columns'], is_numeric($indexName) ? null : $indexName);
+        }
+
         $compositeName = $theJoinTable->getName().'.'.implode('', $localColumns);
         if (isset($addedFks[$compositeName])
             && ($foreignTableName != $addedFks[$compositeName]['foreignTableName']
